/* 
   COPYRIGHT 2006 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - cartographertray.c */
/* rdk@krupczak.org */

/* Windows system tray icon control program */

#include <windows.h>
#include <winsvc.h>
#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <direct.h>

#include "xmp.h"
#include "socklib.h"
#include "xmp_vars.h"
#include "xmp_session.h"
#include "setuplib.h"

/* ************************************************************* */

#define ABOUT_CARTOGRAPHER "\
Cartographer\n\
Cartographertray 0.52\n\
Copyright 2008, Krupczak.org, LLC\n\
\n"

#define THIS_CLASSNAME      "Cartographer System Tray Window"
#define THIS_TITLE          "Cartographer Service"

/* global variable indicating if we are showing a message box or not. */
static BOOL g_bModalState       = FALSE;

/* format of versionStr variable is "KorgVersion<space>%3.2f" 
   that way, our software distribution and updating component
   can correctly figure out version of some distributable
   binary thing */   

/* version 0.51 adds support for SSL */

char *versionStr = "KorgVersion 0.52";

/* stats for xmp communication */
xmpStats_t xmpStats;
int xmp_session_verbose = 0;

/* ************************************************************* */

/* enumerate various numbers we will use; use enumerate to keep them
   unique */
enum {
    //  Tray icon 
    ID_TRAYICON         = 1,

    APPWM_TRAYICON      = WM_APP ,
    APPWM_NOP           = WM_APP + 1,

    //  Our commands
    ID_ABOUT            = 2000,
    ID_STATUS,
    ID_START,
    ID_STOP,
    ID_EXIT,
};

void    AddTrayIcon( HWND hWnd, UINT uID, UINT uCallbackMsg, UINT uIcon, 
                     LPSTR pszToolTip );
void    RemoveTrayIcon( HWND hWnd, UINT uID);
void    ModifyTrayIcon( HWND hWnd, UINT uID, UINT uIcon, LPSTR pszToolTip );

HICON   LoadSmallIcon( HINSTANCE hInstance, UINT uID );

BOOL    ShowPopupMenu( HWND hWnd, POINT *curpos, int wDefaultItem );
void    OnInitMenuPopup( HWND hWnd, HMENU hMenu, UINT uID );

BOOL    OnCommand( HWND hWnd, WORD wID, HWND hCtl );

void    OnTrayIconMouseMove( HWND hWnd );
void    OnTrayIconRBtnUp( HWND hWnd );
void    OnTrayIconLBtnDblClick( HWND hWnd );

void    OnClose( HWND hWnd );

void    RegisterMainWndClass( HINSTANCE hInstance );
void    UnregisterMainWndClass( HINSTANCE hInstance );

int     statusCartographerService();

/* ************************************************************* */

static FILE *fp = NULL;

int logMessage(char *messageStr)
{
     char filename[256];
     char timestr[64];
     char *ptr;
     time_t now;

     if (fp == NULL) {
        sprintf(filename,"c:\\Program Files\\cartographer\\cartographertray.log");
        if ((fp = fopen(filename,"a")) == NULL) {
  	   return -1;
        }
     }

     now = time(NULL);
     ptr = ctime(&now);
     strncpy(timestr,ptr,63);
     timestr[63] = 0;
     if (timestr[strlen(timestr)-1] == '\n')
        timestr[strlen(timestr)-1] = 0;

     if (fp != NULL) {
        fprintf(fp,"%s %s\n",timestr,messageStr);
        fflush(fp);
     }

     return 1;
}

/* ************************************************************* */
/* windows entry point like main() for console apps */

int WINAPI WinMain ( HINSTANCE hInst, HINSTANCE prev, LPSTR cmdline, int show )
{
    HMENU   hSysMenu    = NULL;
    HWND    hWnd        = NULL;
    HWND    hPrev       = NULL;
    MSG     msg;
    WSADATA wsaData;
    char    fqdn[512];
    char    mydir[512];

     if (WSAStartup(0x202,&wsaData) == SOCKET_ERROR) {
        perror("WSAStartup failed");
        WSACleanup();
        exit(-1);
     }

    //  Detect previous instance, and bail if there is one.  
    if ( hPrev = FindWindow( THIS_CLASSNAME, THIS_TITLE ) )
        return 0; //SendMessage ( hPrev, WM_CLOSE, 0, 0 );

    //  We have to have a window, even though we never show it.  This is 
    //  because the tray icon uses window messages to send notifications to 
    //  its owner.  Starting with Windows 2000, you can make some kind of 
    //  "message target" window that just has a message queue and nothing
    //  much else, but we'll be backwardly compatible here.
    RegisterMainWndClass( hInst );

    hWnd = CreateWindow ( THIS_CLASSNAME, THIS_TITLE,
                         0, 0, 0, 100, 100, NULL, NULL, hInst, NULL );

    if ( ! hWnd ) {
        MessageBox ( NULL, "Ack! I can't create the window!", THIS_TITLE, 
                    MB_ICONERROR | MB_OK | MB_TOPMOST );
        return 1;
    }

    /* set our cwd so that we can open files w/o full pathames */
    /* set this up so that we dont have to explicitly state the path */
    if (get_home_dir("cartographertray.exe",mydir,511) > 0) {
       _chdir(mydir);
    }
    else {
       _chdir("c:\\program files\\cartographer");
    }

    /* figure out our FQDN */
    get_fqdn(fqdn,511);

    logMessage("cartographertray starting");
    logMessage(fqdn);

    init_ssl_stuff(fqdn,"cartographertray.pem","cartographertray.csr");

    //  main message loop
    while ( GetMessage ( &msg, NULL, 0, 0 ) ) {
        TranslateMessage ( &msg );
        DispatchMessage ( &msg );
    }

    UnregisterMainWndClass( hInst );

    if (fp != NULL) {
       fclose(fp);
    }

    return msg.wParam;
}

/* ************************************************************* */
/* Add an icon to the system tray. */

void AddTrayIcon( HWND hWnd, UINT uID, UINT uCallbackMsg, UINT uIcon, 
                  LPSTR pszToolTip )
{
    NOTIFYICONDATA   nid;

    memset ( &nid,0,sizeof(nid));

    nid.cbSize              = sizeof( nid );
    nid.hWnd                = hWnd;
    nid.uID                 = uID;
    nid.uFlags              = NIF_ICON | NIF_MESSAGE | NIF_TIP;

    /* set the callback message ID here; it can pretty much be 
       any unique number to our program; this allows us to distinguish
       between events to our icon and events to possibly other windows */
    nid.uCallbackMessage    = uCallbackMsg;

    /* get our icon out of our resources */
    nid.hIcon = LoadSmallIcon(GetModuleHandle( NULL ), uIcon );

#if 0
    /* use this code if we wanted to pull an existing icon out of
       a windows system or other dll/exe */
    {
    char    szIconFile[512];

    GetSystemDirectory ( szIconFile, sizeof( szIconFile ) );
    if ( szIconFile[ strlen( szIconFile ) - 1 ] != '\\' )
        strcat( szIconFile, "\\" );
    strcat( szIconFile, "shell32.dll" );
    //  Icon #23 (0-indexed) in shell32.dll is a "help" icon.
    ExtractIconEx ( szIconFile, 23, NULL, &(nid.hIcon), 1 );
    }
#endif /* 0 */


    strcpy ( nid.szTip, pszToolTip );

    Shell_NotifyIcon ( NIM_ADD, &nid );
}

/* ************************************************************* */
/* if we wanted to modify the icon, here is our chance; we currently
   dont modify the icon */

void ModifyTrayIcon( HWND hWnd, UINT uID, UINT uIcon, LPSTR pszToolTip )
{
    NOTIFYICONDATA  nid;

    memset ( &nid, 0, sizeof( nid ) );

    nid.cbSize  = sizeof( nid );
    nid.hWnd    = hWnd;
    nid.uID     = uID;

    if ( uIcon != (UINT)-1 ) {
        nid.hIcon   = LoadSmallIcon( GetModuleHandle ( NULL ), uIcon );
        nid.uFlags  |= NIF_ICON;
    }

    if ( pszToolTip ) {
        strcpy ( nid.szTip, pszToolTip );
        nid.uFlags  |= NIF_TIP;
    }

    if ( uIcon != (UINT)-1 || pszToolTip )
        Shell_NotifyIcon( NIM_MODIFY, &nid );
}

/* ************************************************************* */
/* Remove an icon from the system tray most likely called on exit;
   we keep the separate in case we want to call this function outside
   of close/exit operation */

void RemoveTrayIcon( HWND hWnd, UINT uID )
{
    NOTIFYICONDATA  nid;

    memset( &nid, 0, sizeof( nid ) );

    nid.cbSize  = sizeof( nid );
    nid.hWnd    = hWnd;
    nid.uID     = uID;

    Shell_NotifyIcon ( NIM_DELETE, &nid );
}

/* ************************************************************* */
/* check the status of the service by trying to connect to it and
   get its version and config info; caller passes in space for
   a ascii string status message; return size of message
 */

int statusCartographerService(char *statusMessage, int size)
{ 
     SC_HANDLE sch,sh;
     SERVICE_STATUS_PROCESS status;
     DWORD bytesNeeded,ret;
     xmpSession_t *s;
     char version[255], sysDescr[255];

     if ((statusMessage == NULL) || (size < 5))
        return -1;

     /* open the service control manager */
     sch = OpenSCManager(NULL,SERVICES_ACTIVE_DATABASE,SC_MANAGER_ALL_ACCESS);
     if (sch == NULL) {
        ret = snprintf(statusMessage,size,
                      "Unable to contact the service control manager");
        return ret;
     }

     /* open the service */
     sh = OpenService(sch,"Cartographer",SERVICE_ALL_ACCESS);
     if (sh == NULL) {
        CloseServiceHandle(sch);
        ret = snprintf(statusMessage,size, 
                       "Unable to open the Cartographer service");
        return ret;
     }

     /* query the service status to see if its running */
     ret = QueryServiceStatusEx(sh,SC_STATUS_PROCESS_INFO,
                                (unsigned char *)&status,
                                sizeof(status),&bytesNeeded);
     if (ret == 0) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        ret = snprintf(statusMessage,size,
                       "Unable to query Cartographer service status");
        return ret;
     }

     if (status.dwCurrentState == SERVICE_STOPPED) {
        CloseServiceHandle(sch);
        CloseServiceHandle(sh);
        ret = snprintf(statusMessage,size,
                       "The Cartographer service is not running");
        return ret;
     }
     CloseServiceHandle(sch);
     CloseServiceHandle(sh);

     /* connect to service via XMP and query */
     if ((s = newXmpSessionConnect(hostname_to_ipaddr("127.0.0.1"),
                                   htons(XMP_PORT),NULL)) == NULL) {
        ret = snprintf(statusMessage,size,
                       "The Cartographer service is running but could not be contacted");
        return ret;
     }
     else {
        /* query for cartographer version, config */
        if ((ret = xmpGetPeerVersion(s,version,sysDescr,255)) > 0) {
	   ret = snprintf(statusMessage,size,"Cartographer %s, %s",
                          version,sysDescr);
        }
        else {
           ret = snprintf(statusMessage,size,
            "The Cartographer service is running but could not be contacted");
        }

        /* done, clean up */
        destroyXmpSession(s);
     }
     
     return ret;
}

/* ************************************************************* */
/* main windows message/event dispatch function */

static LRESULT CALLBACK WindowProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM  lParam )
{
    switch  ( uMsg ) {

        case WM_CREATE:
	    // when we call CreateWindow, we eventually get this message
            // back to our windowproc callback ; we then use this
            // entry point to add our system try icon (rdk)
            AddTrayIcon( hWnd, ID_TRAYICON, APPWM_TRAYICON, 2, THIS_TITLE );
            return  0;

        case APPWM_NOP:
            //  There's a long comment in OnTrayIconRBtnUp() which explains 
            //  what we're doing here.
            return 0;

        //  This is the message which brings tidings of mouse events involving 
        //  our tray icon.  We defined it ourselves.  See AddTrayIcon() for 
        //  details of how we told Windows about it.

        case APPWM_TRAYICON:
            SetForegroundWindow ( hWnd );

            switch ( lParam ) {
                case WM_MOUSEMOVE :
                    OnTrayIconMouseMove( hWnd );
                    return 0;

                case WM_RBUTTONUP :
                    //  There's a long comment in OnTrayIconRBtnUp() which 
                    //  explains what we're doing here.
                    OnTrayIconRBtnUp( hWnd );
                    return 0;

                case WM_LBUTTONDBLCLK :
                    OnTrayIconLBtnDblClick( hWnd );
                    return 0;
            }
            return 0;

        case WM_COMMAND :
	    // this message sent when a user selects a command item from a 
            // a menu . . . 
            return OnCommand( hWnd, LOWORD(wParam), (HWND)lParam );

        case WM_INITMENUPOPUP :
	    // this message sent when a popup menu is about to be come active
            // and it gives us a chance to modify the menu
            OnInitMenuPopup( hWnd, (HMENU)wParam, lParam );
            return 0;

        case WM_CLOSE :
  	    // this message sent when application should terminate
            OnClose( hWnd );
            return DefWindowProc ( hWnd, uMsg, wParam, lParam );

        default:
            return DefWindowProc( hWnd, uMsg, wParam, lParam );
    }
}

/* ************************************************************* */
/* called on exit */

void OnClose( HWND hWnd )
{
    /* Remove icon from system tray. */
    RemoveTrayIcon( hWnd, ID_TRAYICON );

    PostQuitMessage ( 0 );
}

/* ************************************************************* */
/* Create and display our little popupmenu when the user right-clicks 
   on the system tray.
*/

BOOL ShowPopupMenu( HWND hWnd, POINT *curpos, int wDefaultItem )
{
    HMENU   hPop        = NULL;
    int     i           = 0;
    WORD    cmd;
    POINT   pt;

    if ( g_bModalState )
        return FALSE;

    hPop = CreatePopupMenu ();

    if ( ! curpos ) {
        GetCursorPos ( &pt );
        curpos = &pt;
    }

    InsertMenu(hPop,i++,MF_BYPOSITION|MF_STRING,ID_STATUS,
               "Status");

    InsertMenu(hPop,i++,MF_BYPOSITION|MF_SEPARATOR,0,"");

    InsertMenu(hPop,i++,MF_BYPOSITION|MF_STRING,ID_START, 
               "Start Cartographer");
    InsertMenu(hPop,i++,MF_BYPOSITION|MF_STRING,ID_STOP,
               "Stop Cartographer");

    InsertMenu(hPop,i++,MF_BYPOSITION|MF_SEPARATOR,0,"");

    InsertMenu(hPop,i++,MF_BYPOSITION|MF_STRING,ID_ABOUT, 
               "About Cartographer");

    InsertMenu(hPop,i++,MF_BYPOSITION|MF_SEPARATOR,ID_ABOUT,"");

    InsertMenu(hPop,i++,MF_BYPOSITION|MF_STRING,ID_EXIT,"Exit" );

    SetMenuDefaultItem ( hPop, ID_ABOUT, FALSE );

    SetFocus ( hWnd );

    SendMessage ( hWnd, WM_INITMENUPOPUP, (WPARAM)hPop, 0 );

    cmd = TrackPopupMenu ( hPop, TPM_LEFTALIGN | TPM_RIGHTBUTTON 
                            | TPM_RETURNCMD | TPM_NONOTIFY,
                          curpos->x, curpos->y, 0, hWnd, NULL );

    SendMessage( hWnd, WM_COMMAND , cmd, 0 );

    DestroyMenu ( hPop );

    return cmd;
}

/* ************************************************************* */
/* a windows command message came in most likely from menu option */

BOOL OnCommand( HWND hWnd, WORD wID, HWND hCtl )
{
    int ret;
    char message[128];

    if ( g_bModalState )
        return 1;

    //  Have a look at the command and act accordingly
    switch ( wID ) {
        case ID_ABOUT:
           g_bModalState = TRUE;
           MessageBox (hWnd, ABOUT_CARTOGRAPHER, THIS_TITLE, 
                       MB_ICONINFORMATION | MB_OK );
           g_bModalState = FALSE;
           break;

        case ID_START:
	   ret = startService(NULL,0);
           g_bModalState = TRUE;
           switch (ret) {
	     case 1:
	       MessageBox(hWnd,"The Cartographer service was started",
                          THIS_TITLE,
                          MB_OK|MB_ICONINFORMATION);
	        break;

   	     case 2:
	        MessageBox(hWnd,"The Cartographer service is already running",
                           THIS_TITLE,
                           MB_OK|MB_ICONWARNING);
	        break;

	     default:
	        sprintf(message,"Failed to start the Cartographer service, %d",
                        ret);
	        MessageBox(hWnd,message,
                           "Error starting service",
                           MB_OK|MB_ICONERROR);
	        break; 
           }
	   g_bModalState = FALSE;
           break;

        case ID_STOP:
	   ret = stopService(NULL,0);
           g_bModalState = TRUE;
           switch (ret) {
	     case 1:
	       MessageBox(hWnd,"The Cartographer service was stopped",
                          THIS_TITLE,
                          MB_OK|MB_ICONINFORMATION);
	        break;

   	     case 2:
	        MessageBox(hWnd,"The Cartographer service is not running",
                           THIS_TITLE,
                           MB_OK|MB_ICONWARNING);
	        break;

	     default:
	        sprintf(message,"Failed to stop the Cartographer service, %d",
                        ret);
	        MessageBox(hWnd,message,
                           "Error stopping service",
                           MB_OK|MB_ICONERROR);
	        break; 
           }
	   g_bModalState = FALSE;
           break;

        case ID_STATUS:
	   ret = statusCartographerService(message,127);
           g_bModalState = TRUE;
           MessageBox(hWnd,message,THIS_TITLE,
                      MB_OK|MB_ICONINFORMATION);
           g_bModalState = FALSE;
           break;

        case ID_EXIT:
	   PostMessage (hWnd,WM_CLOSE,0,0);
           break;
    }

    return 0;

}

/* ************************************************************* */
/*
  When the mouse pointer drifts over the tray icon, a "tooltip" will be 
  displayed.  Before that happens, we get notified about the movement, so we 
  get a chance to set the "tooltip" text to be something useful.  

*/

void OnTrayIconMouseMove( HWND hWnd )
{
     /* stub - add code here if we want; right now, we're not doing anything
        special */
}

/* ************************************************************* */
/* Right-click on tray icon displays menu */

void OnTrayIconRBtnUp( HWND hWnd )
{
    /*
    This SetForegroundWindow () and PostMessage( , APPWM_NOP, , ) are 
    recommended in some MSDN ; apparently there's a bug in most if 
    not all versions of Windows : Tray icon menus don't vanish properly when 
    cancelled unless you provide special coddling.  

    In MSDN, see: "PRB: Menus for Notification Icons Don't Work Correctly", 
    Q135788:

        mk:@ivt:kb/Source/win32sdk/q135788.htm

    Example code:

        mk:@ivt:pdref/good/code/graphics/gdi/setdisp/c3447_7lbt.htm

    Both of these pseudo-URL's are from the July 1998 MSDN.  Those geniuses 
    have since completely re-broken MSDN at least once, so the pseudo-URL 's 
    are useless with more recent MSDN's.

    In the April 2000 MSDN, you can search titles for "Menus for 
    Notification Icons"; "Don't" in the title has been changed to "Do Not", 
    and searching for either complete title doesn't work anyway.  Good old 
    MSDN.
    */

    SetForegroundWindow( hWnd );

    ShowPopupMenu( hWnd, NULL, -1 );

    PostMessage ( hWnd, APPWM_NOP, 0, 0 );
}

/* ************************************************************* */

void OnTrayIconLBtnDblClick( HWND hWnd )
{
    SendMessage( hWnd, WM_COMMAND, ID_ABOUT, 0 );
}

/* ************************************************************* */

void OnInitMenuPopup( HWND hWnd, HMENU hPop, UINT uID )
{
    //  stub
}

/* ************************************************************* */

void RegisterMainWndClass( HINSTANCE hInstance )
{
    WNDCLASSEX wclx;
    memset( &wclx, 0, sizeof( wclx ) );

    wclx.cbSize         = sizeof ( wclx );
    wclx.style          = 0;
    wclx.lpfnWndProc    = &WindowProc ;
    wclx.cbClsExtra     = 0;
    wclx.cbWndExtra     = 0;
    wclx.hInstance      = hInstance;
    //wclx.hIcon        = LoadIcon ( hInstance, MAKEINTRESOURCE( IDI_TRAYICON ) );
    //wclx.hIconSm      = LoadSmallIcon( hInstance, IDI_TRAYICON );
    wclx.hCursor        = LoadCursor ( NULL, IDC_ARROW );
    wclx.hbrBackground  = (HBRUSH )( COLOR_BTNFACE + 1 );    //  COLOR_* + 1 is
                                                            //  special magic.
    wclx.lpszMenuName   = NULL ;
    wclx.lpszClassName  = THIS_CLASSNAME;

    RegisterClassEx ( &wclx );
}

/* ************************************************************* */

void UnregisterMainWndClass( HINSTANCE hInstance )
{
    UnregisterClass ( THIS_CLASSNAME, hInstance );
}

/* ************************************************************* */

HICON LoadSmallIcon( HINSTANCE hInstance, UINT uID )
{
  return LoadImage ( hInstance, MAKEINTRESOURCE(uID ), IMAGE_ICON, 
                      16, 16, 0 );
}

/* ************************************************************* */
