/* 
   COPYRIGHT 2005 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - xmp_session.h */
/* rdk@krupczak.org */

/* ************************************************************* */
/* internal structures, defs for sessions */

typedef struct xmpSession {
   int sockfd;   /* connected socket */
   int flags;    /* dreaded flags */
   time_t createTime;

   time_t lastActivity;

   /* addressing information (IPv4 only) */
   u_short localPort; 
   u_int   remIp;
   u_short remPort;

   /* user information, permissions; max 16-chars + \0 */
   int xmpVersion;
   char authenUser[18];

   /* per-session stats? */
   pthread_t thr;

   /* send and receive buffer sizes */
   u_char *sendBuf;
   int sendBufSize;
   u_char *recvBuf;
   int recvBufSize;

   /* SSL is connected on top of existing TCP connection
      use void pointers so we dont have to drag around
      ssl include files everywhere 
   */
   void *ssl;
   void *sbio;

} xmpSession_t;

int check_certificate(xmpSession_t *);
int xmpRecvMessage(xmpSession_t *);
int xmpSendMessage(xmpSession_t *, u_char *, int);
int init_ssl_stuff(char *, char *, char *);
int terminate_ssl_stuff();

/* session flags */
#define XMP_SESSION_OUTGOING   0x0001
#define XMP_SESSION_INCOMING   0x0002
#define XMP_SESSION_CONNECTED  0x0004
#define XMP_SESSION_AUTHEN     0x0008

DLLExport int xmpSessionSetAuthenUser(xmpSession_t *, char *);
DLLExport xmpSession_t *newXmpSessionConnect(u_int, u_short, char *);
DLLExport xmpSession_t *newXmpSessionConnectTimeout(u_int, u_short, int, char *);
DLLExport int destroyXmpSession(xmpSession_t *);
DLLExport xmpSession_t *newXmpSession(u_int , u_short, int, int, char *);
DLLExport int isSessionConnected(xmpSession_t *);
DLLExport int xmpSendQueryReceiveReply(xmpSession_t *,xmpPdu_t *,xmpPdu_t *);
DLLExport int xmpGetPeerVersion(xmpSession_t *, char *, char *, int max);
DLLExport int xmpGetPeerPluginList(xmpSession_t *,char *,int);
DLLExport int xmpIsSessionConnected(xmpSession_t *);
DLLExport int xmpIsSessionAuthenticated(xmpSession_t *);
DLLExport int xmpConnectSession(xmpSession_t *);
DLLExport int destroyXmpSession(xmpSession_t *);

DLLExport int xmpSendErrorResponseFatal(xmpSession_t *,int64_t,
                                        xmpErrorStatus_t, u_int);

DLLExport int xmpSendErrorResponse(xmpSession_t *, xmpPdu_t *, 
                                   xmpErrorStatus_t, int errorIndex);

/* ************************************************************* */
